from ..base import _Enum


# Kraken supported symbols
class Symbol(_Enum):
    BCHUSD = 'BCHUSD'
    DASHEUR = 'DASHEUR'
    DASHUSD = 'DASHUSD'
    DASHXBT = 'DASHXBT'
    USDTZUSD = 'USDTUSD'
    XETCXETH = 'ETCETH'
    XETCXXBT = 'ETCXBT'
    XETCZEUR = 'ETCEUR'
    XETCZUSD = 'ETCUSD'
    XETHXXBT = 'ETHXBT'
    # XETHXXBT.d = 'ETHXBT.d'
    XETHZCAD = 'ETHCAD'
    # XETHZCAD.d = 'ETHCAD.d'
    XETHZEUR = 'ETHEUR'
    # XETHZEUR.d = 'ETHEUR.d'
    XETHZGBP = 'ETHGBP'
    # XETHZGBP.d = 'ETHGBP.d'
    XETHZJPY = 'ETHJPY'
    # XETHZJPY.d = 'ETHJPY.d'
    XETHZUSD = 'ETHUSD'
    # XETHZUSD.d = 'ETHUSD.d'
    XICNXETH = 'ICNETH'
    XICNXXBT = 'ICNXBT'
    XLTCXXBT = 'LTCXBT'
    XLTCZEUR = 'LTCEUR'
    XLTCZUSD = 'LTCUSD'
    XMLNXETH = 'MLNETH'
    XMLNXXBT = 'MLNXBT'
    XREPXETH = 'REPETH'
    XREPXXBT = 'REPXBT'
    XREPZEUR = 'REPEUR'
    XREPZUSD = 'REPUSD'
    XXBTZCAD = 'XBTCAD'
    # XXBTZCAD.d = 'XBTCAD.d'
    XXBTZEUR = 'XBTEUR'
    # XXBTZEUR.d = 'XBTEUR.d'
    XXBTZGBP = 'XBTGBP'
    # XXBTZGBP.d = 'XBTGBP.d'
    XXBTZJPY = 'XBTJPY'
    # XXBTZJPY.d = 'XBTJPY.d'
    XXBTZUSD = 'XBTUSD'
    # XXBTZUSD.d = 'XBTUSD.d'
    XXDGXXBT = 'XDGXBT'
    XXLMXXBT = 'XLMXBT'
    XXLMZEUR = 'XLMEUR'
    XXLMZUSD = 'XLMUSD'
    XXMRXXBT = 'XMRXBT'
    XXMRZEUR = 'XMREUR'
    XXMRZUSD = 'XMRUSD'
    XXRPXXBT = 'XRPXBT'
    XZECXXBT = 'ZECXBT'
    XZECZEUR = 'ZECEUR'
    XZECZUSD = 'ZECUSD'


# Kraken supported currencies
class Currency(_Enum):
    DASH = 'DASH'
    KFEE = 'FEE'
    USDT = 'USDT'
    XDAO = 'DAO'
    XETC = 'ETC'
    XETH = 'ETH'
    XICN = 'ICN'
    XLTC = 'LTC'
    XMLN = 'MLN'
    XNMC = 'NMC'
    XREP = 'REP'
    XXBT = 'XBT'
    XXDG = 'XDG'
    XXLM = 'XLM'
    XXMR = 'XMR'
    XXRP = 'XRP'
    XXVN = 'XVN'
    XZEC = 'ZEC'
    ZCAD = 'CAD'
    ZEUR = 'EUR'
    ZGBP = 'GBP'
    ZJPY = 'JPY'
    ZKRW = 'KRW'
    ZUSD = 'USD'



import json as j
import time
from enum import Enum
from json.decoder import JSONDecodeError
from urllib.parse import urljoin

from typing import Iterable

import backoff
import requests
from requests import Response
from requests import Session
from requests.auth import AuthBase

from .common import clean_empty
from .errors import DecodeError
from .errors import InvalidResponse
from .errors import RequestException

TIMEOUT = 30
RETRY_CODES = [
    400, 401, 403, 404, 408, 429,
    500, 502, 503, 504,
]


class Timestamp(object):

    @staticmethod
    def seconds():
        return int(time.time())

    @staticmethod
    def milliseconds():
        return int(time.time() * 1E3)

    @staticmethod
    def microseconds():
        return int(time.time() * 1E6)


timestamp = Timestamp()


class ClientSession(Session):

    def __init__(self,
                 base_url: str,
                 timeout: int=TIMEOUT):
        # Init session
        super().__init__()
        # Instance attributes
        self.auth: AuthBase = None
        self.base_url: str = base_url
        self.timeout: int = timeout
        self.last_nonce: int = 0
        self.last_request_timestamp: int = 0

    def request(self, method, endpoint, *args, **kwargs):
        """Send the request after generating the complete URL."""
        kwargs.setdefault('allow_redirects', True)
        url = self.url_for(endpoint)
        # Clean empty values
        for key in ['data', 'json', 'params']:
            value = kwargs.get(key)
            if value:
                cleaned = clean_empty(value)
                kwargs[key] = cleaned
        return super().request(
            method, url,
            auth=self.auth,
            timeout=self.timeout,
            *args, **kwargs,
        )

    def url_for(self, endpoint: str):
        """Create the URL based off this partial endpoint."""
        return urljoin(self.base_url, endpoint)


class Client(object):
    base_url: str = ''
    error_keys: Iterable[str] = []
    rate_limit: int = 1000  # in milliseconds
    session_cls = ClientSession
    timestamp: Timestamp = timestamp
    retry_codes: Iterable[int] = RETRY_CODES
    # Client defaults
    enable_rate_limit: bool = True
    backoff_factor: float = 1.5  # in seconds
    max_retries: int = 3
    timeout: int = TIMEOUT

    def __init__(self,
                 timeout: int=None,
                 max_retries: int=None,
                 backoff_factor: float=None,
                 enable_rate_limit: bool=None,
                 **kwargs):
        super().__init__(**kwargs)
        # Override defaults
        if timeout is not None:
            self.timeout = timeout
        if max_retries is not None:
            self.max_retries = max_retries
        if enable_rate_limit is not None:
            self.enable_rate_limit = enable_rate_limit
        if backoff_factor is not None:
            self.backoff_factor = backoff_factor
        # Create session
        self.session: ClientSession = self.session_cls(
            self.base_url, self.timeout)
        # Attributes
        self.last_request_timestamp: int = 0

    def get(self, endpoint, **kwargs):
        return self._fetch('GET', endpoint, **kwargs)

    def post(self, endpoint, data=None, json=None, **kwargs):
        return self._fetch('POST', endpoint, data=data, json=json, **kwargs)

    def put(self, endpoint, data=None, **kwargs):
        return self._fetch('PUT', endpoint, data=data, **kwargs)

    def patch(self, endpoint, data=None, **kwargs):
        return self._fetch('PATCH', endpoint, data=data, **kwargs)

    def delete(self, endpoint, **kwargs):
        return self._fetch('DELETE', endpoint, **kwargs)

    def _retry(self, target):
        def give_up_retry(e: RequestException):
            give_up = False
            if e.response:
                code = e.response.status_code
                give_up = code not in self.retry_codes
            return give_up
        return backoff.on_exception(
            backoff.expo,
            RequestException,
            factor=self.backoff_factor,
            max_time=self.timeout,
            max_tries=self.max_retries,
            giveup=give_up_retry,
        )(target)

    def _fetch(self, method, endpoint, *args, **kwargs):
        @self._retry
        def fetch():
            return self._fetch_base(method, endpoint, *args, ** kwargs)
        return fetch()

    def _fetch_base(self, method, endpoint, *args, **kwargs):
        # Rate limit requests
        if self.enable_rate_limit:
            self.throttle()
        self.last_request_timestamp = self.timestamp.milliseconds()
        # Send the request
        response = self.session.request(method, endpoint, *args, **kwargs)
        # Check response for errors
        try:
            response.raise_for_status()
        except requests.HTTPError:
            error_msg = self._get_error_message(response)
            raise InvalidResponse(error_msg, response)
        # Decode the response
        json = self._decode_response(response)
        return json

    def _get_error_message(self, data):
        for error_key in self.error_keys:
            try:
                json = data.json() if isinstance(data, Response) else data
                message = json[error_key]
                if message:
                    msg = j.dumps(message).replace('"', '').rstrip('.')
                    msg = f'{error_key}: {msg}'
                    return msg
            except (JSONDecodeError, KeyError, TypeError):
                return

    def _decode_response(self, response):
        try:
            json = response.json()
        except JSONDecodeError as e:
            error_msg = 'Unable to decode JSON from response (no content)'
            raise DecodeError(error_msg, response) from e
        error_msg = self._get_error_message(json)
        if error_msg:
            raise InvalidResponse(error_msg, response)
        return json

    def throttle(self):
        now = float(self.timestamp.milliseconds())
        elapsed = now - self.last_request_timestamp
        if elapsed < self.rate_limit:
            delay = self.rate_limit - elapsed
            time.sleep(delay / 1E3)

    def __del__(self):
        if self.session:
            self.session.close()


class AuthMixin(object):
    auth_cls = None

    @property
    def auth(self):
        return self.session.auth

    @auth.setter
    def auth(self, auth: AuthBase):
        self.session.auth = auth

    def add_auth(self, *credentials):
        self.auth = self.auth_cls(*credentials)


class ModelMixin(object):
    return_json: bool = False

    def __init__(self, return_json: bool=None):
        super().__init__()
        if return_json:
            self.return_json = return_json


class _Enum(Enum):
    @staticmethod
    def _format_value(value):
        return str(value).upper()

    @classmethod
    def check(cls, value):
        if value is None:
            return value
        if type(value) is cls:
            return value
        try:
            return cls[cls._format_value(value)]
        except KeyError:
            return cls._missing_(value)

    def __str__(self):
        return self.value


class Currency(_Enum):
    @property
    def value(self):
        return super().value['value']

    @property
    def decimals(self):
        return super().value.get('decimals', 2)


class Market(_Enum):
    @staticmethod
    def _format_value(value):
        value = str(value).replace('-', '')
        value = f'{value[:3]}_{value[3:]}'
        return value.upper()

    @property
    def value(self):
        return super().value['value']

    @property
    def base(self):
        return super().value['base']

    @property
    def quote(self):
        return super().value['quote']



from . import constants_v1 as _c1
from . import constants_v2 as _c2
from .client_auth_v1 import BitfinexAuth
from .client_public_v1 import BitfinexPublic
from .client_public_v2 import BitfinexPublic as BitfinexPublicV2

__all__ = [
    'Bitfinex',
    'BitfinexV2',
]


class Bitfinex(object):
    # Enum Types
    Currency = _c1.Currency
    Symbol = _c1.Symbol
    # Clients V1
    Auth = BitfinexAuth
    Public = BitfinexPublic


class BitfinexV2(object):
    # Enum Types
    BookPrecision = _c2.BookPrecision
    Symbol = _c2.Symbol
    # Clients
    # TODO: Implement Bitfinex v2 Auth client
    # Auth = BitfinexAuthV2
    Public = BitfinexPublicV2



from ..base import Client


class BitfinexPublic(Client):
    base_url = 'https://api.bitfinex.com/v1/'
    error_keys = ['message']

    def ticker(self, symbol: str):
        """Gets the innermost bid and asks and information on the most recent trade.

        The ticker is a high level overview of the state of the market. It
        shows you the current best bid and ask, as well as the last trade
        price. It also includes information such as daily volume and how much
        the price has moved over the last day.

        GET https://api.bitfinex.com/v1/pubticker/[symbol]

        Args:
            symbol (str): (Required)
                The symbol you want information about.
                You can find the list of valid symbols by calling the /symbols
                endpoint.

        Returns:
            dict: A dictionary of the following:

            Key         Type                Description
            mid         [float as str]      (bid + ask) / 2
            bid         [float as str]      Innermost bid
            ask         [float as str]      Innermost ask
            last_price  [float as str]      The price at which the last order
                                            executed
            low         [float as str]      Lowest trade price of the last 24h
            high        [float as str]      Highest trade price of the last 24h
            volume      [float as str]      Trading volume of the last 24h
            timestamp   [float as str]      The timestamp at which this
                                            information was valid

        """
        return self.get(f'pubticker/{symbol}')

    def stats(self, symbol: str):
        """Various statistics about the requested pair.

        GET https://api.bitfinex.com/v1/stats/[symbol]

        Args:
            symbol (str): (Required)
                The symbol you want information about.
                You can find the list of valid symbols by calling the /symbols
                endpoint.

        Returns:
            list: A list of dicts with the following:

            Key     Type                Description
            period  [int]               Period covered in days
            volume  [float as str]      Volume in the period

        """
        return self.get(f'stats/{symbol}')

    def today(self, symbol: str):
        """Today's low, high and volume.

        GET https://api.bitfinex.com/v1/today/[symbol]

        Args:
            symbol (str): (Required)
                The symbol you want information about.
                You can find the list of valid symbols by calling the /symbols
                endpoint.

        Returns:
            dict: A dictionary of the following:

            Key     Type                Description
            volume  [float as str]      Today's volume
            high    [float as str]      Today's high price
            low     [float as str]      Today's low price

        """
        return self.get(f'today/{symbol}')

    def lend_book(self,
                  currency: str,
                  limit_bids: int=None,
                  limit_asks: int=None):
        """Get the full margin funding book.

        GET https://api.bitfinex.com/v1/lendbook/[currency]

        Args:
            currency (str): (Required)
                Currency (USD, BTC, etc.)

            limit_bids (int): (Optional, default=50)
                Limit the number of bids returned. May be 0 in which case the
                array of bids is empty.

            limit_asks (int): (Optional, default=50)
                Limit the number of asks returned. May be 0 in which case the
                array of bids is empty.

        Returns:
            dict: A dictionary of the following:

            Key     Type                Description
            bids    [list of dicts]     Array of funding bids
            asks    [list of dicts]     Array of funding asks

            Detail of each dict:

            Key	        Type            Description
            rate        [float as str]  Rate in % per 365 days
            amount      [float as str]
            period      [int]           Minimum period for the margin funding
                                        contract in days
            timestamp   [float as str]
            frr	        [str]           “Yes” if the offer is at Flash Return
                                        Rate, “No” if the offer is at fixed
                                        rate

        """
        return self.get(f'lendbook/{currency}', params={
            'limit_bids': limit_bids,
            'limit_asks': limit_asks,
        })

    def order_book(self,
                   symbol: str,
                   limit_bids: int=None,
                   limit_asks: int=None,
                   group: int=None):
        """Get the full order book.

        GET https://api.bitfinex.com/v1/book/[symbol]

        Args:
            symbol (str): (Required)
                The symbol you want information about.
                You can find the list of valid symbols by calling the /symbols
                endpoint.

            limit_bids (int): (Optional, default=50)
                Limit the number of bids returned. May be 0 in which case the
                array of bids is empty.

            limit_asks (int): (Optional, default=50)
                Limit the number of asks returned. May be 0 in which case the
                array of bids is empty.

            group (int): (Optional, default=1)
                If 1, orders are grouped by price in the order book. If 0,
                orders are sorted individually.

        Returns:
            dict: A dictionary of the following:

            Key     Type                Description
            bids    [list of dicts]     Array of bid orders
            asks    [list of dicts]     Array of ask orders

            Detail of each dict:

            Key         Type            Description
            price       [float as str]
            amount      [float as str]
            timestamp   [float as str]

        """
        return self.get(f'book/{symbol}', params={
            'limit_bids': limit_bids,
            'limit_asks': limit_asks,
            'group': group,
        })

    def trades(self,
               symbol: str,
               timestamp: float=None,
               limit_trades: int=None):
        """Get a list of the most recent trades for the given symbol.

        GET https://api.bitfinex.com/v1/trades/[symbol]

        Args:
            symbol (str): (Required)
                The symbol you want information about.
                You can find the list of valid symbols by calling the /symbols
                endpoint.

            timestamp (float): (Optional)
                Only show trades at or after this timestamp.

            limit_trades (int): (Optional, default=50)
                Limit the number of trades returned. Must be >= 1.

        Returns:
            list: A list of dicts with the following:

            Key         Type            Description
            tid         [int]
            timestamp   [int]
            price       [float as str]
            amount      [float as str]
            exchange    [str]
            type        [str]           'sell' or 'buy'
                                        (can be '' if undetermined)

        """
        return self.get(f'trades/{symbol}', params={
            'timestamp': timestamp,
            'limit_trades': limit_trades,
        })

    def lends(self,
              currency: str,
              timestamp: float=None,
              limit_lends: int=None):
        """Get a list of the most recent lending data for the given currency:

        Total amount lent and rate (in % by 365 days).

        GET https://api.bitfinex.com/v1/lends/[currency]

        Args:
            currency (str): (Required)
                Currency (USD, BTC, etc.)

            timestamp (float): (Optional)
                Only show data at or after this timestamp.

            limit_lends (int): (Optional, default=50)
                Limit the amount of funding data returned. Must be >= 1.

        Returns:
            list: A list of dicts with the following:

            Key         Type	        Description
            rate        [float as str]	Average rate of total funding received
                                        at fixed rates, ie past Flash Return
                                        Rate annualized (% by 365 days)
            amount_lent [float as str]  Total amount of open margin funding in
                                        the given currency
            amount_used [float as str]  Total amount of open margin funding
                                        used in a margin position
                                        in the given currency
            timestamp   [int]

        """
        return self.get(f'lends/{currency}', params={
            'timestamp': timestamp,
            'limit_lends': limit_lends,
        })

    def symbols(self):
        """Get a list of valid symbol IDs.

        GET https://api.bitfinex.com/v1/symbols

        Returns:
            list: A list of symbol names as str.

        """
        return self.get('symbols')

    def symbols_details(self):
        """Get a list of valid symbol IDs and the pair details.

        GET https://api.bitfinex.com/v1/symbols_details

        Returns:
            list: A list of dicts with the following:

            Key                 Type            Description
            pair                [str]           The pair code (symbol)
            price_precision     [int]           Maximum number of significant
                                                digits for price in this pair
            initial_margin      [float as str]  Initial margin required to open
                                                a position in this pair
            minimum_margin      [float as str]  Minimal margin to maintain (%)
            maximum_order_size  [float as str]  Maximum order size of the pair
            expiration          [str]           Expiration date for limited
                                                contracts/pairs

        """
        return self.get('symbols_details')



from ..base import _Enum


class CurrencyPair(_Enum):
    BTC_USD = 'btcusd'
    BTC_EUR = 'btceur'
    EUR_USD = 'eurusd'
    XRP_USD = 'xrpusd'
    XRP_EUR = 'xrpeur'
    XRP_BTC = 'xrpbtc'
    LTC_USD = 'ltcusd'
    LTC_EUR = 'ltceur'
    LTC_BTC = 'ltcbtc'
    ETH_USD = 'ethusd'
    ETH_EUR = 'etheur'
    ETH_BTC = 'ethbtc'
    BCH_USD = 'bchusd'
    BCH_EUR = 'bcheur'
    BCH_BTC = 'bchbtc'


class TimeInterval(_Enum):
    DAY = 'day'
    HOUR = 'hour'
    MINUTE = 'minute'



import math
from collections import namedtuple
from datetime import datetime


def parse_datetime(datetime_str):
    if datetime_str:
        return datetime.strptime(datetime_str, '%Y-%m-%dT%H:%M:%S.%fZ')


def int_or_none(value):
    if value:
        return int(value)


def float_or_none(value):
    if value:
        return float(value)


class Amount(
    namedtuple('amount', [
        'amount',
        'currency',
    ])
):

    @classmethod
    def create_from_json(cls, amount):
        if amount:
            amount = cls(
                amount=float(amount[0]),
                currency=amount[1],
            )
        return amount


class PagesMeta(
    namedtuple('meta', [
        'current_page',
        'total_count',
        'total_pages',
    ])
):

    @classmethod
    def create_from_json(cls, meta):
        if meta:
            return cls(
                current_page=int(meta['current_page']),
                total_count=int(meta['total_count']),
                total_pages=int(meta['total_pages']),
            )
        return meta


class Market(
    namedtuple('market', [
        'id',
        'name',
        'base_currency',
        'quote_currency',
        'minimum_order_amount',
    ]),
):

    @classmethod
    def create_from_json(cls, market):
        return cls(
            id=market['id'],
            name=market['name'],
            base_currency=market['base_currency'],
            quote_currency=market['quote_currency'],
            minimum_order_amount=Amount.create_from_json(
                market['minimum_order_amount']
            ),
        )


class Ticker(
    namedtuple('ticker', [
        'last_price',
        'min_ask',
        'max_bid',
        'volume',
        'price_variation_24h',
        'price_variation_7d'
    ])
):

    @classmethod
    def create_from_json(cls, ticker):
        return cls(
            last_price=Amount.create_from_json(ticker['last_price']),
            min_ask=Amount.create_from_json(ticker['min_ask']),
            max_bid=Amount.create_from_json(ticker['max_bid']),
            volume=Amount.create_from_json(ticker['volume']),
            price_variation_24h=float_or_none(ticker['price_variation_24h']),
            price_variation_7d=float_or_none(ticker['price_variation_7d']),
        )


class Quotation(
    namedtuple('quotation', [
        'amount',
        'base_balance_change',
        'base_exchanged',
        'fee',
        'incomplete',
        'limit',
        'order_amount',
        'quote_balance_change',
        'quote_exchanged',
        'type',
    ])
):

    @classmethod
    def create_from_json(cls, quotation):
        return cls(
            amount=Amount.create_from_json(
                quotation['amount']),
            base_balance_change=Amount.create_from_json(
                quotation['base_balance_change']),
            base_exchanged=Amount.create_from_json(
                quotation['base_exchanged']),
            fee=Amount.create_from_json(
                quotation['fee']),
            incomplete=quotation['incomplete'],
            limit=Amount.create_from_json(
                quotation['limit']),
            order_amount=Amount.create_from_json(
                quotation['order_amount']),
            quote_balance_change=Amount.create_from_json(
                quotation['quote_balance_change']),
            quote_exchanged=Amount.create_from_json(
                quotation['quote_exchanged']),
            type=quotation['type'],
        )


class OrderBookEntry(
    namedtuple('book_entry', [
        'price',
        'amount',
    ])
):

    @classmethod
    def create_from_json(cls, book_entry):
        return cls(
            price=float(book_entry[0]),
            amount=float(book_entry[1]),
        )


class OrderBook(
    namedtuple('order_book', [
        'asks',
        'bids',
    ])
):

    @classmethod
    def create_from_json(cls, order_book):
        return cls(
            asks=[OrderBookEntry.create_from_json(entry)
                  for entry in order_book['asks']],
            bids=[OrderBookEntry.create_from_json(entry)
                  for entry in order_book['bids']],
        )


class FeePercentage(
    namedtuple('fee_percentage', [
        'value',
    ])
):

    @classmethod
    def create_from_json(cls, fee_percentage):
        return cls(
            value=float(fee_percentage['value']),
        )


class Balance(
    namedtuple('balance', [
        'id',
        'account_id',
        'amount',
        'available_amount',
        'frozen_amount',
        'pending_withdraw_amount',
    ])
):

    @classmethod
    def create_from_json(cls, balance):
        return cls(
            id=balance['id'],
            account_id=balance['account_id'],
            amount=Amount.create_from_json(
                balance['amount']),
            available_amount=Amount.create_from_json(
                balance['available_amount']),
            frozen_amount=Amount.create_from_json(
                balance['frozen_amount']),
            pending_withdraw_amount=Amount.create_from_json(
                balance['pending_withdraw_amount']),
        )


class Order(
    namedtuple('order', [
        'id',
        'account_id',
        'amount',
        'created_at',
        'fee_currency',
        'limit',
        'market_id',
        'original_amount',
        'paid_fee',
        'price_type',
        'state',
        'total_exchanged',
        'traded_amount',
        'type',
    ])
):

    @classmethod
    def create_from_json(cls, order):
        return cls(
            id=order['id'],
            account_id=order['account_id'],
            amount=Amount.create_from_json(order['amount']),
            created_at=parse_datetime(order['created_at']),
            fee_currency=order['fee_currency'],
            limit=Amount.create_from_json(order['limit']),
            market_id=order['market_id'],
            original_amount=Amount.create_from_json(order['original_amount']),
            paid_fee=Amount.create_from_json(order['paid_fee']),
            price_type=order['price_type'],
            state=order['state'],
            total_exchanged=Amount.create_from_json(order['total_exchanged']),
            traded_amount=Amount.create_from_json(order['traded_amount']),
            type=order['type'],
        )


class OrderPages(
    namedtuple('order_pages', [
        'orders',
        'meta',
    ])
):

    @classmethod
    def create_from_json(cls, orders, pages_meta):
        return cls(
            orders=[Order.create_from_json(order)
                    for order in orders],
            meta=PagesMeta.create_from_json(pages_meta),
        )


class BalanceEvent(
    namedtuple('balance_event', [
        'id',
        'account_id',
        'created_at',
        'currency',
        'event',
        'event_ids',
        'new_amount',
        'new_available_amount',
        'new_frozen_amount',
        'new_frozen_for_fee',
        'new_pending_withdraw_amount',
        'old_amount',
        'old_available_amount',
        'old_frozen_amount',
        'old_frozen_for_fee',
        'old_pending_withdraw_amount',
        'transaction_type',
        'transfer_description',
    ])
):

    @classmethod
    def create_from_json(cls, event):
        return cls(
            id=event['id'],
            account_id=event['account_id'],
            created_at=parse_datetime(event['created_at']),
            currency=event['currency'],
            event=event['event'],
            event_ids=event['event_ids'],
            new_amount=event['new_amount'],
            new_available_amount=event['new_available_amount'],
            new_frozen_amount=event['new_frozen_amount'],
            new_frozen_for_fee=event['new_frozen_for_fee'],
            new_pending_withdraw_amount=event['new_pending_withdraw_amount'],
            old_amount=event['old_amount'],
            old_available_amount=event['old_available_amount'],
            old_frozen_amount=event['old_frozen_amount'],
            old_frozen_for_fee=event['old_frozen_for_fee'],
            old_pending_withdraw_amount=event['old_pending_withdraw_amount'],
            transaction_type=event['transaction_type'],
            transfer_description=event['transfer_description'],
        )


class BalanceEventPages(
    namedtuple('event_pages', [
        'balance_events',
        'meta',
    ])
):

    @classmethod
    def create_from_json(cls, events, total_count, page):
        return cls(
            balance_events=[BalanceEvent.create_from_json(event)
                            for event in events],
            meta=PagesMeta(
                current_page=page or 1,
                total_count=total_count,
                total_pages=math.ceil(total_count / len(events))),
        )


class TradeTransaction(
    namedtuple('trade_transaction', [
        'id',
        'market_id',
        'created_at',
        'updated_at',
        'amount_sold',
        'price_paid',
        'ask_order',
        'bid_order',
        'triggering_order',
    ])
):

    @classmethod
    def create_from_json(cls, transaction):
        return cls(
            id=transaction['id'],
            market_id=transaction['market_id'],
            created_at=parse_datetime(transaction['created_at']),
            updated_at=parse_datetime(transaction['updated_at']),
            amount_sold=Amount.create_from_json(
                [transaction['amount_sold'] / 1e8,
                 transaction['amount_sold_currency']]),
            price_paid=Amount.create_from_json(
                [transaction['price_paid'] / 1e2,
                 transaction['price_paid_currency']]),
            ask_order=Order.create_from_json(transaction['ask']),
            bid_order=Order.create_from_json(transaction['bid']),
            triggering_order=Order.create_from_json(
                transaction['triggering_order']),
        )


class TradeTransactionPages(
    namedtuple('trade_transaction_pages', [
        'trade_transactions',
        'meta',
    ])
):

    @classmethod
    def create_from_json(cls, transactions, pages_meta):
        return cls(
            trade_transactions=[TradeTransaction.create_from_json(transaction)
                                for transaction in transactions],
            meta=PagesMeta.create_from_json(pages_meta),
        )


class TradeEntry(
    namedtuple('trade_transaction_pages', [
        'timestamp',
        'amount',
        'price',
        'direction',
    ])
):

    @classmethod
    def create_from_json(cls, entry):
        return cls(
            timestamp=int(entry[0]),
            amount=float(entry[1]),
            price=float(entry[2]),
            direction=entry[3],
        )


class Trades(
    namedtuple('trade_transaction_pages', [
        'timestamp',
        'last_timestamp',
        'entries'
    ])
):

    @classmethod
    def create_from_json(cls, trades):
        return cls(
            timestamp=int_or_none(trades['timestamp']),
            last_timestamp=int_or_none(trades['last_timestamp']),
            entries=[TradeEntry.create_from_json(entry)
                     for entry in trades['entries']],
        )


class TransferData(
    namedtuple('transfer', [
        'type',
        'address',
        'tx_hash',
    ])
):

    @classmethod
    def create_from_json(cls, transfer, address_key):
        if transfer:
            return cls(
                type=transfer['type'],
                address=transfer.get(address_key),
                tx_hash=transfer.get('tx_hash'),
            )


class Transfer(
    namedtuple('transfer', [
        'id',
        'created_at',
        # 'updated_at', Missing from response?
        'amount',
        'fee',
        'currency',
        'state',
        'data',
    ])
):
    address_key = None
    data_key = None

    @classmethod
    def create_from_json(cls, transfer):
        return cls(
            id=transfer['id'],
            created_at=parse_datetime(transfer.get('created_at')),
            amount=Amount.create_from_json(transfer['amount']),
            # Fee is only returned on withdrawals
            fee=Amount.create_from_json(transfer.get('fee')),
            currency=transfer['currency'],
            state=transfer['state'],
            data=TransferData.create_from_json(
                transfer[cls.data_key], cls.address_key),
        )


class Withdrawal(Transfer):
    address_key = 'target_address'
    data_key = 'withdrawal_data'


class Deposit(Transfer):
    address_key = 'address'
    data_key = 'deposit_data'


class WithdrawalPages(
    namedtuple('withdrawal_pages', [
        'withdrawals',
        'meta',
    ])
):

    @classmethod
    def create_from_json(cls, withdrawals, pages_meta):
        return cls(
            withdrawals=[Withdrawal.create_from_json(withdrawal)
                         for withdrawal in withdrawals],
            meta=PagesMeta.create_from_json(pages_meta),
        )


class DepositPages(
    namedtuple('deposit_pages', [
        'deposits',
        'meta',
    ])
):

    @classmethod
    def create_from_json(cls, deposits, pages_meta):
        return cls(
            deposits=[Deposit.create_from_json(deposit)
                      for deposit in deposits],
            meta=PagesMeta.create_from_json(pages_meta),
        )


class AveragePrice(
    namedtuple('reports', [
        'datetime',
        'amount'
    ])
):

    @classmethod
    def create_from_json(cls, report):

        return cls(
            datetime=report[0],
            amount=report[1]
        )


class Candlestick(
    namedtuple('report', [
        'datetime',
        'open',
        'high',
        'low',
        'close',
        'volume'
    ])
):

    @classmethod
    def create_from_json(cls, report):

        return cls(
            datetime=report[0],
            open=report[1],
            high=report[2],
            low=report[3],
            close=report[4],
            volume=report[5],
        )



import base64
import hashlib
import hmac

from requests import PreparedRequest as P

from .client_public import KrakenPublic
from ..auth import HMACAuth
from ..base import AuthMixin


class KrakenHMACAuth(HMACAuth):

    api_key_header = 'API-Key'
    signature_header = 'API-Sign'
    algorithm = 'sha512'

    def add_nonce(self, r: P, nonce: str):
        body = self.parse_data(r.body)
        body['nonce'] = nonce
        r.prepare_body(data=body, files=None)

    def build_message(self, r: P, nonce: str):
        auth = (nonce + r.body).encode()
        digest = hashlib.sha256(auth).digest()
        message = r.path_url.encode() + digest
        return message

    def sign(self, msg: str):

        h = hmac.new(key=base64.b64decode(self.secret),
                     msg=msg,
                     digestmod=self.algorithm)

        signature = base64.b64encode(h.digest()).decode()

        return signature


class KrakenAuth(KrakenPublic, AuthMixin):
    auth_cls = KrakenHMACAuth

    def __init__(self, 
                 key: str,
                 secret: str,
                 timeout: int=None,
                 max_retries: int=None,
                 backoff_factor: float=None,
                 enable_rate_limit: bool=None):
        super().__init__(timeout, max_retries, backoff_factor,
                         enable_rate_limit)
        self.add_auth(key, secret)

    # Private user data -------------------------------------------------------
    # Get account balance.
    def balance(self):
        return self.post('private/Balance')

    # Get trade balance.
    def trade_balance(self,
                      asset: str=None,
                      asset_class: str=None):
        return self.post('private/TradeBalance', data={
            'asset': str(asset) if asset else None,
            'aclass': str(asset_class) if asset_class else None,
        })

    # Get open orders.
    def open_orders(self,
                    include_trades: bool=None,
                    userref: str=None):
        return self.post('private/OpenOrders', data={
            'trades': include_trades,
            'userref': userref,
        })

    # Get closed orders.
    def closed_orders(self,
                      include_trades: bool=None,
                      userref: str=None,
                      start: int=None,
                      end: int=None,
                      ofs: int=None,
                      closetime: int=None):
        return self.post('private/ClosedOrders', data={
            'trades': include_trades,
            'userref': userref,
            'start': start,
            'end': end,
            'ofs': ofs,
            'closetime': closetime,
        })

    # Query orders info.
    def query_orders(self,
                     txid: list,
                     include_trades: bool=None,
                     userref: str=None):
        return self.post('private/QueryOrders', data={
            'txid': txid,
            'trades': include_trades,
            'userref': userref,
        })

    # Get trades history.
    def trades_history(self,
                       trade_type: str=None,
                       include_trades: bool=None,
                       start: int=None,
                       end: int=None,
                       ofs: int=None):
        return self.post('private/TradesHistory', data={
            'type': str(trade_type) if trade_type else None,
            'trades': include_trades,
            'start': start,
            'end': end,
            'ofs': ofs,
        })

    # Query trades info.
    def query_trades(self,
                     txid: list,
                     include_trades: bool=None):
        return self.post('private/QueryTrades', data={
            'txid': txid,
            'trades': include_trades,
        })

    # Query trades info.
    def open_positions(self,
                       txid: list=None,
                       include_pl: bool=None):
        return self.post('private/OpenPositions', data={
            'txid': txid,
            'docalcs': include_pl,
        })

    # Get ledgers info.
    def ledgers(self,
                asset_class: str=None,
                asset: str=None,
                ledger_type: str=None,
                start: int=None,
                end: int=None,
                ofs: int=None):
        return self.post('private/Ledgers', data={
            'aclass': str(asset_class) if asset_class else None,
            'asset': str(asset) if asset else None,
            'type': str(ledger_type) if ledger_type else None,
            'start': start,
            'end': end,
            'ofs': ofs,
        })

    # Query ledgers.
    def query_ledgers(self, ledger_id: str):
        return self.post('private/QueryLedgers', data={
            'id': ledger_id,
        })

    # Get trade volume.
    def trade_volume(self,
                     pair: str=None,
                     fee_info: bool=None):
        return self.post('private/TradeVolume', data={
            'pair': str(pair) if pair else None,
            'fee-info': fee_info,
        })

    # Private user trading  ---------------------------------------------------
    # Add standard order
    def add_order(self,
                  pair: str,
                  direction: str,
                  order_type: str,
                  volume: float,
                  price: float=None,
                  price2: float=None,
                  leverage: float=None,
                  oflags: list=None,
                  starttm: int=None,
                  expiretm: int=None,
                  userref: str=None,
                  validate: bool=None,
                  c_ordertype: str=None,
                  c_price: float=None,
                  c_price2: float=None):
        return self.post('private/AddOrder', data={
            'pair': str(pair),
            'type': str(direction),
            'ordertype': str(order_type),
            'price': price,
            'price2': price2,
            'volume': volume,
            'leverage': leverage,
            'oflags': oflags,
            'starttm': starttm,
            'expiretm': expiretm,
            'userref': userref,
            'validate': validate,
            'close[ordertype]': str(c_ordertype) if order_type else None,
            'close[price]': c_price,
            'close[price2]': c_price2,
        })

    # Cancel open order
    def cancel_order(self, txid: str):
        return self.post('private/CancelOrder', data={
            'txid': txid,
        })

    # Private user funding  ---------------------------------------------------
    # Get deposit methods.
    def deposit_methods(self,
                        asset: str,
                        asset_class: str=None):
        return self.post('private/DepositMethods', data={
            'asset': str(asset),
            'aclass': str(asset_class) if asset_class else None,
        })

    # Get deposit addresses
    def deposit_addresses(self,
                          asset: str,
                          method: str,
                          asset_class: str=None,
                          new: bool=None):
        return self.post('private/DepositAddresses', data={
            'asset': str(asset),
            'method': str(method),
            'aclass': str(asset_class) if asset_class else None,
            'new': new,
        })

    # Get status of recent deposits
    def deposit_status(self,
                       asset: str,
                       method: str,
                       asset_class: str=None):
        return self.post('private/DepositStatus', data={
            'asset': str(asset),
            'method': str(method),
            'aclass': str(asset_class) if asset_class else None,
        })

    # Get withdrawal information
    def withdraw_info(self,
                      asset: str,
                      amount: float,
                      key: str,
                      asset_class: str=None):
        return self.post('private/WithdrawInfo', data={
            'asset': str(asset),
            'aclass': str(asset_class) if asset_class else None,
            'amount': amount,
            'key': key,
        })

    # Withdraw funds
    def withdraw(self,
                 asset: str,
                 amount: float,
                 key: str,
                 asset_class: str=None):
        return self.post('private/Withdraw', data={
            'asset': str(asset),
            'aclass': str(asset_class) if asset_class else None,
            'amount': amount,
            'key': key,
        })

    # Get status of recent withdrawals
    def withdraw_status(self,
                        asset: str,
                        method: str,
                        asset_class: str=None):
        return self.post('private/WithdrawStatus', data={
            'asset': str(asset),
            'method': str(method),
            'aclass': str(asset_class) if asset_class else None,
        })

    # Request withdrawal cancellation
    def withdraw_cancel(self,
                        asset: str,
                        refid: str,
                        asset_class: str=None):
        return self.post('private/WithdrawCancel', data={
            'asset': str(asset),
            'aclass': str(asset_class) if asset_class else None,
            'refid': refid,
        })



from ..auth import ApiKeyAuth
from ..base import AuthMixin, Client
from ..common import format_date_iso


class CurrencyLayer(Client, AuthMixin):
    """CurrencyLayer API Client

    Documentation:
    https://currencylayer.com/documentation
    """
    # https is only enabled for paid subscriptions
    base_url = 'http://apilayer.net/api/'
    error_keys = ['error']
    timeout = 120

    def __init__(self,
                 access_key: str,
                 timeout: int=None,
                 max_retries: int=None,
                 backoff_factor: float=None,
                 enable_rate_limit: bool=None):
        super().__init__(timeout, max_retries, backoff_factor,
                         enable_rate_limit)
        self.auth = ApiKeyAuth(access_key, api_key_param='access_key')

    def currencies(self):
        """Returns all currently supported currencies.

        Returns:
            dict:
                {
                    "success": true,
                    "terms": "https://currencylayer.com/terms",
                    "privacy": "https://currencylayer.com/privacy",
                    "currencies": {
                        "AED": "United Arab Emirates Dirham",
                        "AFN": "Afghan Afghani",
                        "ALL": "Albanian Lek",
                        "AMD": "Armenian Dram",
                        "ANG": "Netherlands Antillean Guilder",
                        ...
                    }
                }
        """
        return self.get('list')

    def live_rates(self,
                   base: str=None,
                   currencies: list=None):
        """Return real-time exchange rates.

        REQUIRED PLAN: FREE

        Exchange rate data is refreshed every 60 minutes for Free & Basic
        plans, every 10 minutes for Professional, and every 60 seconds for
        Enterprise.

        Both free and paid users may access real-time exchange rates.
        Optionally, it is possible to define an additional base currency,
        or specific output currencies using the currencies parameter.

        Args:
            base (str):
                REQUIRED PLAN: BASIC
                Request exchange rates relative to a different base currency
                (default is USD).
            currencies (list):
                Limit the request to a set of specific currencies codes.

        Returns:
            dict:
                {
                    "success": true,
                    "terms": "https://currencylayer.com/terms",
                    "privacy": "https://currencylayer.com/privacy",
                    "timestamp": 1432400348,
                    "source": "USD",
                    "quotes": {
                        "USDAUD": 1.278342,
                        "USDEUR": 0.908019,
                        "USDGBP": 0.645558,
                        "USDPLN": 3.731504,
                        ...
                    }
                }
        """
        return self.get('live', base, currencies)

    def historical(self,
                   date_for,
                   base: str=None, 
                   currencies: list=None):
        """Return historical exchange rates for the date requested.

        REQUIRED PLAN: FREE

        Args:
            date_for: (required)
                Date to return exchange rates for. Accepts `datetime` and `str`
                in ISO format (2000-01-01).
            base (str):
                REQUIRED PLAN: BASIC
                Request exchange rates relative to a different base currency
                (default is USD).
            currencies (list):
                Limit the request to a set of specific currencies codes.

        Returns:
            dict:
                {
                    "success": true,
                    "terms": "https://currencylayer.com/terms",
                    "privacy": "https://currencylayer.com/privacy",
                    "historical": true,
                    "date": "2005-02-01",
                    "timestamp": 1107302399,
                    "source": "USD",
                    "quotes": {
                        "USDAED": 3.67266,
                        "USDALL": 96.848753,
                        "USDAMD": 475.798297,
                        "USDANG": 1.790403,
                        "USDARS": 2.918969,
                        "USDAUD": 1.293878,
                        ...
                    }
                }
        """
        return self.get('historical', base, currencies, params={
            'date': format_date_iso(date_for),
        })

    def time_frame(self,
                   start: str,
                   end: str,
                   base: str=None,
                   currencies: list=None):
        """Returns daily historical exchange rates for a time-period.

        (maximum range: 365 days)

        REQUIRED PLAN: PRO

        Args:
            start: (required)
                Start date for the requested time-frame. Accepts `datetime`
                and `str` in ISO format (2000-01-01).
            end: (required)
                End date for the requested time-frame. Accepts `datetime`
                and `str` in ISO format (2000-01-01).
            base (str):
                REQUIRED PLAN: BASIC
                Request exchange rates relative to a different base currency
                (default is USD).
            currencies (list):
                Limit the request to a set of specific currencies codes.

        Returns:
            dict:
                {
                    "success": true,
                    "terms": "https://currencylayer.com/terms",
                    "privacy": "https://currencylayer.com/privacy",
                    "timeframe": true,
                    "start_date": "2010-03-01",
                    "end_date": "2010-04-01",
                    "source": "USD",
                    "quotes": {
                        "2010-03-01": {
                            "USDUSD": 1,
                            "USDGBP": 0.668525,
                            "USDEUR": 0.738541,
                        },
                        "2010-03-02": {
                            "USDUSD": 1,
                            "USDGBP": 0.668827,
                            "USDEUR": 0.736145,
                        },
                        ...
                    },
                }
        """
        return self.get('timeframe', base, currencies, params={
            'start_date': format_date_iso(start),
            'end_date': format_date_iso(end),
        })

    def convert(self,
                amount: float,
                from_currency: str,
                to_currency: str,
                date_for: str=None):
        """Perform a currency conversion for the amount.

        Converts from a `from` currency, to a `to` currency at the latest
        exchange rate. If date_for is passed, the rate for that date will be
        used.

        REQUIRED PLAN: BASIC

        Args:
            amount (float): (required)
                Amount to convert.
            from_currency (str): (required)
                Currency to convert from (the amount's currency).
            to_currency (str): (required)
                Currency to convert to (the return currency).
            date_for:
                End date for the requested time-frame. Accepts `datetime`
                and `str` in ISO format (2000-01-01).
        """
        return self.get('convert', params={
            'amount': amount,
            'from': from_currency,
            'to': to_currency,
            'date': format_date_iso(date_for),
        })

    def change(self,
               base: str=None,
               currencies: list=None):
        """Return the change (margin and percentage) from yesterday's EOD (End Of Day).

        REQUIRED PLAN: ENTERPRISE

        Args:
            base (str):
                REQUIRED PLAN: BASIC
                Request exchange rates relative to a different base currency
                (default is USD).
            currencies (list):
                Limit the request to a set of specific currencies codes.

        Returns:
            dict:
                {
                    "success": true,
                    "terms": "https://currencylayer.com/terms",
                    "privacy": "https://currencylayer.com/privacy",
                    "change": true,
                    "start_date": "2005-01-01",
                    "end_date": "2010-01-01",
                    "source": "USD",
                    "quotes": {
                        "USDAUD": {
                            "start_rate": 1.281236,
                            "end_rate": 1.108609,
                            "change": -0.1726,
                            "change_pct": -13.4735
                        },
                        "USDEUR": {
                            "start_rate": 0.73618,
                            "end_rate": 0.697253,
                            "change": -0.0389,
                            "change_pct": -5.2877
                        },
                        "USDMXN": {
                            "start_rate": 11.149362,
                            "end_rate": 13.108757,
                            "change": 1.9594,
                            "change_pct": 17.5741,
                        },
                        ...
                    },
                }
        """
        return self.get('change', base, currencies)

    def change_time_frame(self,
                          start: str,
                          end: str,
                          base: str=None,
                          currencies: list=None):
        """Return the change (margin and percentage) for a specific time-frame.

        REQUIRED PLAN: ENTERPRISE

        Args:
            start: (required)
                Start date for the requested time-frame. Accepts `datetime`
                and `str` in ISO format (2000-01-01).
            end: (required)
                End date for the requested time-frame. Accepts `datetime`
                and `str` in ISO format (2000-01-01).
            base (str):
                Request exchange rates relative to a different base currency
                (default is USD).
            currencies (list):
                Limit the request to a set of specific currencies codes.

        Returns:
            dict:
                {
                    "success": true,
                    "terms": "https://currencylayer.com/terms",
                    "privacy": "https://currencylayer.com/privacy",
                    "change": true,
                    "start_date": "2005-01-01",
                    "end_date": "2010-01-01",
                    "source": "USD",
                    "quotes": {
                        "USDAUD": {
                            "start_rate": 1.281236,
                            "end_rate": 1.108609,
                            "change": -0.1726,
                            "change_pct": -13.4735
                        },
                        "USDEUR": {
                            "start_rate": 0.73618,
                            "end_rate": 0.697253,
                            "change": -0.0389,
                            "change_pct": -5.2877
                        },
                        "USDMXN": {
                            "start_rate": 11.149362,
                            "end_rate": 13.108757,
                            "change": 1.9594,
                            "change_pct": 17.5741,
                        },
                        ...
                    },
                }
        """
        return self.get('change', base, currencies, params={
            'start': format_date_iso(start),
            'end': format_date_iso(end),
        })

    # OVERRIDES ---------------------------------------------------------------
    def get(self,
            endpoint: str,
            base: str=None,
            currencies: list=None,
            params: dict=None):
        params = params or {}
        if base is not None:
            params['source'] = base
        if isinstance(currencies, list) or isinstance(currencies, tuple):
            currencies = ','.join(currencies)
        if currencies is not None:
            params['currencies'] = currencies
        return super().get(endpoint, params=params)
